--=|The Database Diagram.|=--

--  unique_class_members    unique_member_names  
--                    1      1
--                    +--+   |
--                       |   |
--    unique_class_names |   |     unique_return_type
--                    1  |   |      1
--                    |  |   |      |
--                    |  |   |      M
--    wk_class_flags  |  |   |     unique_signatures  
--                0   |  |   |      1
--                |   |  |   |      |
--                |   |  |   M      M
--  wk_components |   |  |  unique_member_signatures
--             1  |   |  |   1
--             |  |   |  | +-+
--             M  M   M  | |
--            wk_classes | |  wk_member_types 
--             0  M   1  | |   0        
--             +--+   |  | | +-+  
--                    M  M M M      
--                   wk_members  - - - >  unique_keywords

--=||=--
-- ===========================================================================
-- ===========================================================================
--    C R E A T E S
-- ===========================================================================
-- ===========================================================================

--=|Create the tables.|=--

    CREATE TABLE IF NOT EXISTS
      wk_class_flags
        (
          code                          TEXT NOT NULL PRIMARY KEY,
          meaning                       TEXT,
          description                   TEXT
        )
;
    CREATE TABLE IF NOT EXISTS
      wk_member_types
        (
          code                          TEXT NOT NULL PRIMARY KEY,
          meaning                       TEXT,
          sort_key                      INTEGER
        )
;
    CREATE TABLE IF NOT EXISTS
      wk_components
        ( 
          id                            INTEGER PRIMARY KEY,
          name                          TEXT NOT NULL,
          comment_count                 INTEGER
        )
;
    CREATE TABLE IF NOT EXISTS
      wk_classes
        ( 
          id                            INTEGER PRIMARY KEY,
          name                          TEXT NOT NULL,
          inherits_name                 TEXT,
          inherits_id                   INTEGER,

          wk_components_fk              INTEGER NOT NULL,
          wk_class_flags_fk          TEXT,

          FOREIGN KEY ( wk_components_fk )
            REFERENCES  wk_components( id )
              ON UPDATE CASCADE
              ON DELETE CASCADE,

          FOREIGN KEY ( wk_class_flags_fk )
            REFERENCES  wk_class_flags( code )
              ON UPDATE CASCADE
              ON DELETE CASCADE
        )
;
    CREATE INDEX IF NOT EXISTS wk_components_fk___wk_classes
               ON  wk_classes( wk_components_fk )
;
    CREATE INDEX wk_class_flags_fk___wk_classes
              ON wk_classes( wk_class_flags_fk )
;
    CREATE INDEX wk_classes__inherits_id
              ON wk_classes( inherits_id )
;
    CREATE TABLE IF NOT EXISTS
      wk_members
        (
          id                            INTEGER PRIMARY KEY,
          name                          TEXT NOT NULL,
          return_type                   TEXT,

          wk_classes_fk                 INTEGER NOT NULL,
          wk_member_types_fk            TEXT,

          FOREIGN KEY ( wk_classes_fk )
            REFERENCES  wk_classes( id )
              ON UPDATE CASCADE
              ON DELETE CASCADE,

          FOREIGN KEY ( wk_member_types_fk )
            REFERENCES  wk_member_types( code )
              ON UPDATE CASCADE
              ON DELETE CASCADE
        )
;
    CREATE INDEX IF NOT EXISTS 
                   wk_classes_fk___wk_members
    ON wk_members( wk_classes_fk )
;
    CREATE INDEX IF NOT EXISTS
                   wk_member_types_fk___wk_members
    ON wk_members( wk_member_types_fk )
;



--=|Create the loading tables.|=--

    CREATE TABLE IF NOT EXISTS
      load_class_flags
        ( 
          code          TEXT,
          meaning       TEXT,
          description   TEXT
        )
;
    CREATE TABLE IF NOT EXISTS
      load_member_types
        ( 
          code          TEXT,
          meaning       TEXT,
          sort_key      INTEGER
        )
;
    CREATE TABLE IF NOT EXISTS
      load_components
        ( 
          name          TEXT NOT NULL,
          comment_count INTEGER
        )
;
    CREATE TABLE IF NOT EXISTS
      load_classes
        ( 
          wk_components_name   TEXT NOT NULL,
          name                 TEXT NOT NULL,
          inherits_name        TEXT,
          inherits_id          INTEGER,
          wk_class_flags_fk TEXT,
          wk_components_fk     INTEGER
        )

;
    CREATE TABLE IF NOT EXISTS
      look_classes
        ( 
          wk_components_name  TEXT,
          wk_classes_name     TEXT,
          wk_classes_id       INTEGER
        )
;
    CREATE TABLE IF NOT EXISTS
      load_members
        ( 
          wk_components_name  TEXT,
          wk_classes_name     TEXT,
          name                TEXT,
          wk_member_types_fk  TEXT,
          return_type         TEXT,

          wk_classes_fk       INTEGER
        )

;
--=||=--
-- ===========================================================================
-- ===========================================================================
--    D R O P S
-- ===========================================================================
-- ===========================================================================

--=|Drop the loading tables.|=--

    DROP TABLE IF EXISTS load_class_flags;
    DROP TABLE IF EXISTS load_member_types;
    DROP TABLE IF EXISTS load_components;
    DROP TABLE IF EXISTS load_classes;
    DROP TABLE IF EXISTS look_classes;
    DROP TABLE IF EXISTS load_members;

--=||=--
-- ===========================================================================
-- ===========================================================================
--    U P D A T E S
-- ===========================================================================
-- ===========================================================================

--=|Update the matching Components.|=--

    UPDATE 
      wk_components
        SET comment_count = 
          (
            SELECT load_components.comment_count
            FROM   load_components
            WHERE  load_components.name = wk_components.name
          )
        WHERE EXISTS 
          (
            SELECT load_components.comment_count
            FROM   load_components
            WHERE  load_components.name          =  wk_components.name
              AND  load_components.comment_count <> wk_components.comment_count
              AND  load_components.comment_count IS NOT NULL
          )
;
--=|Update the matching Classes.|=--

--=||=--
-- ===========================================================================
-- ===========================================================================
--    S E T   R E F E R E N C E   K E Y S
-- ===========================================================================
-- ===========================================================================

--=|Set the loaded Classes foreign key to Components.|=--

    UPDATE load_classes
    SET    wk_components_fk = 
           (
              SELECT wk_components.id 
              FROM   wk_components 
              WHERE  wk_components.name = load_classes.wk_components_name
           ) 
;
--=|Set the loaded Members foreign key to Classes.|=--

    UPDATE load_members
    SET    wk_classes_fk = 
           (
              SELECT look_classes.wk_classes_id 
              FROM   look_classes 
              WHERE  look_classes.wk_classes_name    = load_members.wk_classes_name 
              AND    look_classes.wk_components_name = load_members.wk_components_name
           )
;

--=|Set the inherits key in Classes.|=--

    UPDATE wk_classes
    SET    inherits_id = 
          (
             SELECT A.id
             FROM      wk_classes 
             AS     A
             WHERE  A.name              = wk_classes.inherits_name 
             AND    A.wk_components_fk  = wk_classes.wk_components_fk 
          )
;


--=||=--
-- ===========================================================================
-- ===========================================================================
--    A P P E N D S
-- ===========================================================================
-- ===========================================================================

--=|Append the new Class Flags.|=--

    INSERT INTO 
      wk_class_flags
        ( 
          code, meaning, description
        )
        SELECT    load_class_flags.code,
                  load_class_flags.meaning,
                  load_class_flags.description
        FROM      load_class_flags
        LEFT JOIN                         wk_class_flags 
        ON        load_class_flags.code = wk_class_flags.code
        WHERE                             wk_class_flags.code IS NULL        
;
--=|Append the new Member Types.|=--

    INSERT INTO 
      wk_member_types
        ( 
          code, meaning, sort_key
        )
        SELECT    load_member_types.code,
                  load_member_types.meaning,
                  load_member_types.sort_key
        FROM      load_member_types
        LEFT JOIN                          wk_member_types 
        ON        load_member_types.code = wk_member_types.code
        WHERE                              wk_member_types.code IS NULL        
;
--=|Append the new Components.|=--

    INSERT INTO 
      wk_components
        ( 
          name, comment_count
        )
        SELECT    load_components.name,
                  load_components.comment_count
        FROM      load_components 
        LEFT JOIN                        wk_components 
        ON        load_components.name = wk_components.name
        WHERE                            wk_components.name IS NULL        
;
--=|Append missing Components from Classes.|=--

    INSERT INTO 
      wk_components( name, comment_count )
        
        SELECT      load_classes.wk_components_name AS name,
                    -1                              AS comment_count

        FROM        load_classes 
        LEFT JOIN                                     wk_components
        ON          load_classes.wk_components_name = wk_components.name
        GROUP BY                                      wk_components.name, 
                                                      wk_components.comment_count
        HAVING                                        wk_components.name IS NULL
;
--=|Append the new Classes.|=--

    INSERT INTO 
      wk_classes
        ( 
          name, inherits_name, wk_class_flags_fk, wk_components_fk
        )
        SELECT     load_classes.name,
                   load_classes.inherits_name,
                   load_classes.wk_class_flags_fk,
                   load_classes.wk_components_fk
        FROM       load_classes
        LEFT JOIN                                  wk_classes
        ON         load_classes.wk_components_fk = wk_classes.wk_components_fk
        WHERE                                      wk_classes.id IS NULL        

;
--=|Append the new Members.|=--

    INSERT INTO 
      wk_members
        ( 
          name, return_type, wk_member_types_fk, wk_classes_fk
        )
        SELECT     load_members.name,
                   load_members.return_type,
                   load_members.wk_member_types_fk,
                   load_members.wk_classes_fk
        FROM       load_members 
        LEFT JOIN                               wk_members 
        ON         load_members.name          = wk_members.name
        AND        load_members.wk_classes_fk = wk_members.wk_classes_fk
        WHERE                                   wk_members.id IS NULL        

;
--=||=--
-- ===========================================================================
-- ===========================================================================
--    M A K E S
-- ===========================================================================
-- ===========================================================================


--=|Make a Componet/Classes to Class ID lookup table.|=--
    
    INSERT INTO 
      look_classes ( wk_components_name, wk_classes_name, wk_classes_id )
      
        SELECT     wk_components.name AS wk_components_name, 
                   wk_classes.name    AS wk_classes_name, 
                   wk_classes.id      AS wk_classes_id
        FROM                                     wk_components 
        INNER JOIN wk_classes
        ON         wk_classes.wk_components_fk = wk_components.id
;

--=||=--
-- ===========================================================================
-- ===========================================================================
--    U N I Q U E    T A B L E   C R E A T I O N
-- ===========================================================================
-- ===========================================================================


--=|Make the Unique Member Signatures table.|=--

    DROP TABLE IF EXISTS unique_member_signatures;
    DROP INDEX IF EXISTS unique_member_signatures__name;
    DROP INDEX IF EXISTS unique_member_signatures__return_type;
    DROP INDEX IF EXISTS unique_member_signatures__member_type;
  
    
    CREATE TABLE unique_member_signatures 
    AS
    SELECT   wk_members.name                AS name,
             wk_members.return_type           AS return_type,
             wk_members.wk_member_types_fk  AS member_type,
             count(id)                      AS member_count
    
    FROM     wk_members
    
    GROUP BY name,
             return_type,
             member_type
    
    ORDER BY name,
             return_type,
             member_type
;


    CREATE INDEX unique_member_signatures__name
              ON unique_member_signatures( name )
;
    CREATE INDEX unique_member_signatures__return_type
              ON unique_member_signatures( return_type )
;
    CREATE INDEX unique_member_signatures__member_type
              ON unique_member_signatures( member_type )
;
--=|Make the Unique Member Names table.|=--

    DROP TABLE IF EXISTS unique_member_names;
    DROP INDEX IF EXISTS unique_member_names__name;

    
    CREATE TABLE unique_member_names
    AS
    SELECT   name              AS name,
             count(name)       AS signature_count,
             sum(member_count) AS member_count
    
    FROM     unique_member_signatures
    
    GROUP BY name
    ORDER BY name
;


    CREATE INDEX unique_member_names__name
              ON unique_member_names( name )
;
--=|Make the Unique Signatures table.|=--

    DROP TABLE IF EXISTS unique_signatures;
    DROP INDEX IF EXISTS unique_signatures__return_type;


    CREATE TABLE unique_signatures
    AS
    SELECT                                return_type,
                                          member_type,

             count( member_type )      AS name_count,
             sum( member_count )       AS member_count
    
    FROM     unique_member_signatures
    
    GROUP BY return_type, member_type

    ORDER BY return_type, member_type

;
    CREATE INDEX unique_signatures__return_type
              ON unique_signatures( return_type )
;

--=|Make the Unique Arguments table.|=--

    DROP TABLE IF EXISTS unique_return_type;
    DROP INDEX IF EXISTS unique_return_type__return_type;


    CREATE TABLE unique_return_type
    AS
    SELECT   return_type                 AS return_type,
             count( member_type )      AS member_type_count,
             sum( name_count )         AS name_count,
             sum( member_count )       AS member_count
    
    FROM     unique_signatures
    
    GROUP BY return_type
    ORDER BY return_type
;

    CREATE INDEX unique_return_type__return_type
              ON unique_return_type( return_type )
;



--=|Make the Unique Class Names table.|=--

    DROP TABLE IF EXISTS unique_class_names;
    DROP INDEX IF EXISTS unique_class_names__class_name;

    CREATE TABLE unique_class_names
    AS

    SELECT     wk_classes.name                        AS class_name,
               count( wk_classes.wk_components_fk )   AS component_count

    FROM       wk_classes
            
    GROUP BY   class_name

    ORDER BY   class_name
;

    CREATE INDEX unique_class_names__class_name
              ON unique_class_names( class_name )
;

--=|Make the Unique Class Members table.|=--

    DROP TABLE IF EXISTS unique_class_members;
    DROP INDEX IF EXISTS unique_class_members__class_name;
    DROP INDEX IF EXISTS unique_class_members__member_name;

    CREATE TABLE unique_class_members
    AS

    SELECT     wk_classes.name                        AS class_name,
               wk_members.name                        AS member_name,
               count( wk_members.id )                 AS count

    FROM       wk_members
    INNER JOIN                            wk_classes
    ON         wk_members.wk_classes_fk = wk_classes.id
            
    GROUP BY   class_name, member_name

    ORDER BY   class_name, member_name
;
    CREATE INDEX unique_class_members__class_name
              ON unique_class_members( class_name )
;
    CREATE INDEX unique_class_members__member_name
              ON unique_class_members( member_name )
;

--=|Make the Unique Keywords table.|=--

    DROP TABLE IF EXISTS unique_keywords;
    DROP INDEX IF EXISTS unique_keywords__keyword;


    CREATE TABLE unique_keywords
    AS

    SELECT     wk_members.name                        AS keyword

    FROM       wk_members
    INNER JOIN                            wk_classes
    ON         wk_members.wk_classes_fk = wk_classes.id

    WHERE      wk_classes.name = "."
  
    GROUP BY   keyword

    ORDER BY   keyword
;

    CREATE INDEX unique_keywords__keyword
              ON unique_keywords( keyword )
;

--=||=--
-- ===========================================================================
-- ===========================================================================
--     S E L E C T I O N S
-- ===========================================================================
-- ===========================================================================

--=|Classes with Components|=--

SELECT wk_classes.name          AS ClassName,
       wk_components.name       AS ComponentName,
       wk_classes.inherits_name AS InheritsName

FROM       wk_classes 
INNER JOIN                               wk_components 
ON         wk_classes.wk_components_fk = wk_components.id

ORDER BY   ClassName, ComponentName	 
;

--=|Class/Component/Member|=--

    SELECT     wk_classes.name               AS ClassName,
               wk_classes.inherits_name      AS InheritsName,
               wk_classes.wk_class_flags_fk  AS ClassFlags,
               wk_components.name            AS ComponentName,
               wk_members.wk_member_types_fk AS MemberType,
               wk_members.name               AS MemberName,
               wk_members.return_type        AS ReturnType

    FROM       wk_components

	INNER JOIN                    wk_classes
    ON         wk_components.id = wk_classes.wk_components_fk

    INNER JOIN                                    wk_members                                 
    ON                            wk_classes.id = wk_members.wk_classes_fk

    ORDER BY   ClassName, ClassFlags, ComponentName, MemberType, MemberName

--=|Member Types|=--

    SELECT   wk_member_types.code,
             wk_member_types.meaning
    FROM     wk_member_types
    ORDER BY wk_member_types.sort_key

--=|Class Flags|=--

    SELECT   wk_class_flags.code,
             wk_class_flags.meaning
    FROM     wk_class_flags


